/* This file is part of the KDE project
   Copyright (C) 2002 Simon Hausmann <hausmann@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "commands.h"
#include "window.h"
#include "frontend.h"
#include "events.h"

#include <qtimer.h>
#include <qvariant.h>

#include <klineeditdlg.h>
#include <klocale.h>

using namespace IRC;

ReplyHandler::ReplyHandler( Window *parent, const char *name )
    : QObject( parent, name ), m_window( parent ), m_frontend( 0 )
{
    if ( parent )
        m_frontend = &parent->frontend();
}

void ReplyHandler::finished()
{
    emit replyHandlerFinished( this );
    deleteLater();
}

Command::Command( Window *parent, const char *name )
    : ReplyHandler( parent, name ), m_finished( false )
{
}

void Command::sendMessage( const Message &msg )
{
    emit sendMessage( msg, this );
}

void Command::finished()
{
    if ( m_finished )
        return;
    m_finished = true;

    emit commandFinished( this );
    ReplyHandler::finished();
}

MacroCommand::MacroCommand( Window *parent, const char *name )
    : Command( parent, name )
{
}

MacroCommand::MacroCommand( const Message &messageToSend, Window *parent, const char *name )
    : Command( parent, name )
{
    m_messages << messageToSend;
}

MacroCommand::MacroCommand( const MessageList &messagesToSend, Window *parent, const char *name )
    : Command( parent, name ), m_messages( messagesToSend )
{
}

void MacroCommand::start()
{
    MessageList::ConstIterator it = m_messages.begin();
    for (; it != m_messages.end(); ++it )
        sendMessage( *it );
}

SimpleCommand::SimpleCommand( Window *parent, const char *name )
    : MacroCommand( parent, name )
{
}

SimpleCommand::SimpleCommand( ReplyHandler *replyHandler, Window *parent, const char *name )
    : MacroCommand( parent, name )
{
    setReplyHandler( replyHandler );
}

SimpleCommand::SimpleCommand( const Message &messageToSend, ReplyHandler *replyHandler, Window *parent, const char *name )
    : MacroCommand( messageToSend, parent, name )
{
    setReplyHandler( replyHandler );
}

SimpleCommand::SimpleCommand( const MessageList &messagesToSend, ReplyHandler *replyHandler, Window *parent, const char *name )
    : MacroCommand( messagesToSend, parent, name )
{
    setReplyHandler( replyHandler );
}

void SimpleCommand::handleNumericMessage( const IRC::Message &msg )
{
    emit internalRedirect( msg );
}

bool SimpleCommand::usesTerminatingCommandSequence() const
{
    if ( !m_replyHandler )
        return false;

    QVariant result = m_replyHandler->property( "usesTerminatingCommandSequence" );
    return result.isValid() && result.toBool();
}

void SimpleCommand::setReplyHandler( ReplyHandler *replyHandler )
{
    connect( this, SIGNAL( internalRedirect( const IRC::Message & ) ),
             replyHandler, SLOT( handleNumericMessage( const IRC::Message & ) ) );

    connect( replyHandler, SIGNAL( replyHandlerFinished( IRC::ReplyHandler * ) ),
             this, SLOT( finished() ) );
    
    m_replyHandler = replyHandler;
}

void SimpleCommand::customEvent( QCustomEvent *ev )
{
    if ( !MessageReceiverRegistrationEvent::test( ev ) ) {
        MacroCommand::customEvent( ev );
        return;
    }

    MessageReceiverRegistrationEvent *evt = static_cast<MessageReceiverRegistrationEvent *>( ev );
    if ( evt->unregisteredHandler() ) 
        finished();
}

LoginCommand::LoginCommand( const QString &host, Window *parent, const char *name )
    : MacroCommand( parent, name ), m_connInfo( parent->frontend().connectionInfo() )
{
    if ( !m_connInfo.password().isEmpty() ) {
        PassMessage msg;
        msg.setPassword( m_connInfo.password() );
        m_messages << msg;
    }

    NickMessage nmsg;
    nmsg.setNick( m_connInfo.nick() );
    m_messages << nmsg;
 
    UserMessage umsg;
    umsg.setUser( m_connInfo.user() );
    umsg.setHost( host );
    umsg.setServer( m_connInfo.host() );
    umsg.setRealName( m_connInfo.realName() );
    m_messages << umsg;
}

void LoginCommand::handleNumericMessage( const IRC::Message &msg )
{
    switch ( msg.numericType() ) {
        case ERR_NICKNAMEINUSE: {
            bool ok = false;

            QString alternateNick = KLineEditDlg::getText( i18n( "Nick Name Already In Use" ),
                                                           i18n( "Nick '%1' is already in use. Please choose another nick name:" ).arg( m_connInfo.nick() ),
                                                           QString::null, &ok );

            // ### eek, must transfer to frontend
            m_connInfo.setNick( alternateNick );
            NickMessage nmsg;
            nmsg.setNick( alternateNick );
            sendMessage( nmsg );

            break;
        }
        case RPL_WELCOME: {
            qDebug( "WELCOME!" );
            // it's not really unhandled, but this way the user sees the message.
            m_frontend->unhandledMessage( msg );
            finished();
            break;
        }
        default: break;
    }
}

DefaultReplyHandler::DefaultReplyHandler( Window *parent, const char *name )
    : ReplyHandler( parent, name )
{
}

void DefaultReplyHandler::handleNumericMessage( const IRC::Message &msg )
{
    QStringList params = msg.parameters();
    params.pop_front();
    QString message = "[" + msg.prefix() + "] " + params.join( " " );
    m_window->printMessage( message );
}

NamesReplyHandler::NamesReplyHandler( const QStringList &channels, Window *parent, const char *name )
    : DefaultReplyHandler( parent, name ), m_channels( channels )
{
}

NamesReplyHandler::NamesReplyHandler( const QString &channel, Window *parent, const char *name )
    : DefaultReplyHandler( parent, name )
{
    m_channels << channel;
}

void NamesReplyHandler::handleNumericMessage( const IRC::Message &msg )
{
    switch ( msg.numericType() ) {
        case RPL_NAMREPLY: {
            parseNamesReply( msg );
            break;
        }
        case RPL_ENDOFNAMES: {
            QString channel = msg.parameter( 1 ).lower();
            m_channels.remove( channel );

            if ( m_channels.isEmpty() )
                emitInfoAndFinish();
            break;
        }
        default: DefaultReplyHandler::handleNumericMessage( msg ); break;
    }
}

void NamesReplyHandler::parseNamesReply( const IRC::Message &msg )
{
    ChannelInfo channel;
    int channelIndex = 1;
    int nicksIndex = 2;

    if ( msg.parameter( 1 )[ 0 ] != '#' ) {
        QChar typeChar = msg.parameter( 1 )[ 0 ];

        if ( typeChar == '@' )
            channel.type = SecretChannel;
        else if ( typeChar == '*' )
            channel.type = PrivateChannel;
        else /* if ( typeChar == '=' ) */
            channel.type = PublicChannel;

        channelIndex++;
        nicksIndex++;
    }

    channel.name = msg.parameter( channelIndex ).lower();
    QStringList nicks = QStringList::split( " ", msg.parameter( nicksIndex ) );

    insertNicksIntoMap( channel, nicks );
}

void NamesReplyHandler::insertNicksIntoMap( const ChannelInfo &channel, const QStringList &nicks )
{
    QStringList::ConstIterator it = nicks.begin();
    for ( ; it != nicks.end(); ++it ) {
        QString nick = *it;

        NickInfo info;
        if ( nick[ 0 ] == '@' ) {
            nick.remove( 0, 1 );
            info.isOp = true;
        } else
            info.isOp = false;

        info.name = nick;

        qDebug( "INSERT %s(%s) FOR %s", info.name.ascii(), info.isOp ? "OP" : "NOOP", channel.name.ascii() );
        m_users[ channel ] << info;
    }
}

void NamesReplyHandler::emitInfoAndFinish()
{
    emit channelUsers( m_users );
    finished();
}

NamesCommand::NamesCommand( const QStringList &channels, Window *parent, const char *name )
    : SimpleCommand( parent, name )
{
    NamesMessage msg;
    msg.setChannels( channels );
    m_messages << msg;

    NamesReplyHandler *handler = new NamesReplyHandler( channels, parent, name );
    setReplyHandler( handler );

    connect( handler, SIGNAL( channelUsers( const IRC::NamesReplyHandler::ChannelUsersMap & ) ),
             this, SIGNAL( channelUsers( const IRC::NamesReplyHandler::ChannelUsersMap & ) ) );
}

JoinCommand::JoinCommand( const JoinMessage::ChannelInfoList &channelInfo, Window *parent, const char *name )
    : NamesCommand( channelInfo.channelList(), parent, name )
{
    // remove NAMES command
    m_messages.clear();

    JoinMessage msg;
    msg.setChannelInfo( channelInfo );
    m_messages << msg;
}

void JoinCommand::handleNumericMessage( const IRC::Message &msg )
{
    /*
    if ( msg.numericType() == RPL_TOPIC ) {
        QString channel = msg.parameter( 1 ).lower();
        QString topic = msg.parameter( 2 );

        Window *wnd = m_frontend->windowForChannel( channel, Frontend::ReturnNullIfNotExisting );

        if ( wnd ) 
            wnd->setTitle( topic );

        return;
    }
    */

    NamesCommand::handleNumericMessage( msg );
}

UserCommand::UserCommand( const Message &message, Window *parent, const char *name )
    : SimpleCommand( message, new DefaultReplyHandler( parent ), parent, name )
{
}

WhoisReplyHandler::WhoisReplyHandler( Window *parent, const char *name )
    : DefaultReplyHandler( parent, name )
{
}

void WhoisReplyHandler::handleNumericMessage( const IRC::Message &msg )
{
    switch ( msg.numericType() ) {
        case RPL_WHOISUSER:
            parseUser( msg );
            break;
        case RPL_WHOISCHANNELS:
            parseChannels( msg );
            break;
        case RPL_WHOISSERVER: {
                QString serverHostName = msg.parameter( 2 );
                QString serverInfo = msg.parameter( 3 );

                m_user.setIRCServer( ServerInfo( serverHostName, serverInfo ) );
                break;
            }
        case RPL_WHOISOPERATOR:
            m_user.setIRCOperator( true );
            break;
        case RPL_AWAY:
            m_user.setAwayMessage( msg.parameter( 2 ) );
            break;
        case RPL_WHOISIDLE:
            m_user.setIdleTime( msg.parameter( 2 ).toInt() );
            break;
        case RPL_WHOISIDENTIFIED:
            // ### TODO (add attribute to Person)
            break;
        case RPL_ENDOFWHOIS:
            qDebug( "USER: %s", m_user.toWhoisString().ascii() );
            qDebug( "Away Msg: %s", m_user.awayMessage().ascii() );
            qDebug( "isIRCOp: %s", ( m_user.isIRCOperator() ? "yes" : "no" ) );
            qDebug( "IDLETIME: %i", m_user.idleTime() );
            qDebug( "on server %s (%s)", m_user.ircServer().hostName.ascii(), m_user.ircServer().description.ascii() );


            /*
            for ( ChannelInfoList::ConstIterator it = m_channels.begin();
                  it != m_channels.end(); ++it )
                qDebug( "ON CHANNEL: %s (isOp: %s, hasVoice: %s)",
                        ( *it ).name.ascii(),
                        ( ( *it ).isOp ? "yes" : "no" ),
                        ( ( *it ).hasVoice ? "yes" : "no" ) );
            */

            emit userDetailsAvailable( m_user );
            finished();
            break;
        default: DefaultReplyHandler::handleNumericMessage( msg ); break;
    }
}

void WhoisReplyHandler::parseUser( const IRC::Message &msg )
{
    m_user.setNick( msg.parameter( 1 ) );
    m_user.setUser( msg.parameter( 2 ) );
    m_user.setHost( msg.parameter( 3 ) );
    m_user.setRealName( msg.parameter( 5 ) );
}

void WhoisReplyHandler::parseChannels( const IRC::Message &msg )
{
    QStringList rawChannelList = QStringList::split( " ", msg.parameter( 2 ) );
    for ( QStringList::ConstIterator it = rawChannelList.begin();
          it != rawChannelList.end(); ++it ) {

        ChannelUserAttributes attributes;

        QString channelName;

        QChar flags = ( *it )[ 0 ];

        if ( flags != '#' ) {

            if ( flags == '@' )
                attributes.isChannelOperator = true;
            else if ( flags == '+' )
                attributes.hasVoice = true;

            channelName = ( *it ).mid( 1 ).lower();
        } else
            channelName = ( *it ).lower();

        m_user.channels()[ channelName ] = attributes;
    }
}
 
#include "commands.moc"

/* vim: et sw=4 ts=4
 */
